#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Functions for saving and loading Pyomo objects to json
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# We need to work with some Pyomo internals for this
# pylint: disable=protected-access

import json
import datetime
import time
import gzip
import logging

from pyomo.environ import (
    Param,
    Var,
    BooleanVar,
    Expression,
    Block,
    Constraint,
    Suffix,
    value,
)
from pyomo.core.base.param import ParamData
from pyomo.core.base.component import ComponentData

_log = logging.getLogger(__name__)

# Some more information about this module
__author__ = "John Eslick"
__format_version__ = 4


def _can_serialize(o):
    try:
        json.dumps(o)
        return True
    except TypeError:
        return False


def _set_active(o, d):
    """
    Set if component is active, used for read active attribute callback.
    Args:
        o: object whose attribute is to be set
        d: attribute value
    Returns:
        None
    """
    try:
        if d:
            o.activate()
        else:
            o.deactivate()
    except AttributeError:
        # if the object doesn't have activate and deactivate that's fine just
        # keep going
        pass


def _set_fixed(o, d):
    """
    Set if variable is fixed, used for read fixed attribute callback.
    Args:
        o: object whose attribute is to be set
        d: attribute value
    Returns:
        None
    """
    if d:
        o.fix()
    else:
        o.unfix()


def _get_value(o):
    """
    Get object value attribute callback.
    Args:
        o: object whose attribute is to be set
        d: attribute value
    Returns:
        None
    """
    return value(o, exception=False)


def _set_value(o, d):
    """
    Set object value attribute callback. This doesn't allow the value of an
    immutable parameter to be set (which would raise an exception in Pyomo)
    Args:
        o: object whose attribute is to be set
        d: attribute value
    Returns:
        None
    """
    if isinstance(o, Param) and not o._mutable:
        return  # ignore requests to set immutable params
    else:
        try:
            o.value = d
        except AttributeError:
            o = d  # this would be an indexed parameter


def _set_lb(o, d):
    """
    Set variable lower bound, used for read lb attribute callback.
    Args:
        o: object whose attribute is to be set
        d: attribute value
    Returns:
        None
    """
    o.setlb(d)


def _set_ub(o, d):
    """
    Set variable upper bound, use for read ub attribute callback.
    Args:
        o: object whose attribute is to be set
        d: attribute value
    Returns:
        None
    """
    o.setub(d)


def _value_if_not_fixed(o, d):
    """
    Returns a list of attributes to read for a variable, only whether it is
    fixed for non-fixed variables and if it is fixed and the value for fixed
    variables.  The allows you to set up a serializer that only reads fixed
    variable values.

    Args:
        o: Pyomo component being loaded
        d: State dictionary for the component o.
    Returns:
        An attribute list to read. Loads fixed for either fixed or un-fixed
        variables, but only reads in values for unfixed variables.  This is
        useful for initialization functions.
    """
    if o.fixed:
        return ()
    return ("value",)


def _only_fixed(o, d):
    """
    Returns a list of attributes to read for a variable, only whether it is
    fixed for non-fixed variables and if it is fixed and the value for fixed
    variables.  The allows you to set up a serializer that only reads fixed
    variable values.

    Args:
        o: Pyomo component being loaded
        d: State dictionary for the component o.
    Returns:
        An attribute list to read. Loads fixed for either fixed or un-fixed
        variables, but only reads in values for unfixed variables.  This is
        useful for initialization functions.
    """
    if d["fixed"]:
        return ("value", "fixed")
    else:
        return ("fixed",)


class Counter(object):
    """
    This is a counter object, which is an easy way to pass an integer pointer
    around between methods.
    """

    def __init__(self):
        self.count = 0


class StoreSpec(object):
    """
    A StoreSpec object tells the serializer functions what to read or write.
    The default settings will produce a StoreSpec configured to load/save the
    typical attributes required to load/save a model state.

    Args:
        classes: Dictionary of classes to read/write. The class is the dictionary
            key and the value is a list or tuple with two elements 1) a list or
            tuple of attributes to read/write and 2) a filter function. The
            attributes can be an empty list and the filter function can be None.
            Filter functions if present only apply to reading components and do
            not affect writing. Filter functions takes two arguments. The first
            is the object being read to and the second is the state dictionary
            being read from. The state dictionary contains keys for the
            attributes that were written. Based on the state of the component
            and the stored state of the component, the filter function returns
            a list of attributes to read.  For example, a filter function can be
            used to only read the values of variables that were fixed when they
            were written. The order of the class keys is important and should
            go from specific classes to general ones. For example, if a
            UnitModel is a Block, but you want to store extra attributes for a
            UnitModel, UnitModel should come before Block in the key set.
        data_classes: This takes the same form as the classes argument, the
            classes should be component data types.
        ignore_missing: If True will ignore a component or attribute that exists
            in the model, but not in the stored state. If false an exception
            will be raised for things in the model that should be loaded but
            aren't in the stored state. Extra items in the stored state will not
            raise an exception regardless of this argument.
        suffix: If True store suffixes and component ids.  If false, don't store
            suffixes. This is a legacy option. The preferred way to store a Suffix
            of not is just to include Suffix in classes or not.
        suffix_filter: None to store all suffixes or a list of suffixes to store.
    """

    def __init__(
        self,
        classes=None,
        data_classes=None,
        ignore_missing=True,
        suffix=None,
        suffix_filter=None,
    ):
        if classes is None:
            classes = {
                Param: (("_mutable",), None),
                Var: ((), None),
                BooleanVar: ((), None),
                Expression: ((), None),
                Block: (("active",), None),
                Constraint: (("active",), None),
                Suffix: ((), None),
            }

        if data_classes is None:
            data_classes = {
                Var._ComponentDataClass: (
                    ("fixed", "stale", "value", "lb", "ub"),
                    None,
                ),
                BooleanVar._ComponentDataClass: (("fixed", "stale", "value"), None),
                ParamData: (("value",), None),
                int: (("value",), None),
                float: (("value",), None),
                str: (("value",), None),
                Expression._ComponentDataClass: ((), None),
                Block._ComponentDataClass: (("active",), None),
                Constraint._ComponentDataClass: (("active",), None),
            }

        # convert old style list/tuple classes arg to dict if needed
        if isinstance(classes, (list, tuple)):
            self.classes = {}
            for c in classes:
                self.classes[c[0]] = c[1:]
        else:
            self.classes = classes
        # convert old style list/tuple data_classes arg to dict if needed
        if isinstance(data_classes, (list, tuple)):
            self.data_classes = {}
            for c in data_classes:
                self.data_classes[c[0]] = c[1:]
        else:
            self.data_classes = data_classes
        # Callbacks are used for attributes that cannot be directly get or set
        self.write_cbs = {  # Write callbacks (writing state so get attr)
            "value": _get_value
        }
        self.read_cbs = {  # Reads callbacks (reading in state so set attr)
            "_mutable": None,
            "active": _set_active,
            "fixed": _set_fixed,
            "lb": _set_lb,
            "ub": _set_ub,
            "value": _set_value,
        }
        # Block and BlockData are required for model structure
        if Block not in self.classes:
            self.classes[Block] = ((), None)
        if Block._ComponentDataClass not in data_classes:
            self.data_classes[Block._ComponentDataClass] = ((), None)
        # If suffix is None, decide by whether in classes, else add or remove
        # suffix based on option.  May deprecate the suffix option.
        if suffix is not None:
            if not suffix and Suffix in classes:
                del self.classes[Suffix]
            elif suffix and Suffix not in classes:
                self.classes[Suffix] = ((), None)
        # Create filter function lists, use None if not supplied
        for i, c in self.classes.items():
            if len(c) < 2:
                self.classes[i] = (c[0], None)
        for i, c in self.data_classes.items():
            if len(c) < 2:
                self.data_classes[i] = (c[0], None)
        self.ignore_missing = ignore_missing
        self.suffix_filter = suffix_filter

    def set_read_callback(self, attr, cb=None):
        """
        Set a callback to set an attribute, when reading from json or dict.
        """
        self.read_cbs[attr] = cb

    def set_write_callback(self, attr, cb=None):
        """
        Set a callback to get an attribute, when writing to json or dict.
        """
        self.write_cbs[attr] = cb

    def get_class_attr_list(self, o):
        """
        Look up what attributes to save/load for an Component object.
        Args:
            o: Object to look up attribute list for.
        Return:
            A list of attributes and a filter function for object type
        """
        alist = None  # Attributes to store
        ff = None  # Load filter function
        for cl in self.classes:
            if isinstance(o, cl):
                alist = self.classes[cl][0]
                ff = self.classes[cl][1]
                break
        if isinstance(o, Block._ComponentDataClass):
            # If you're here you are trying to serialize an element of an
            # indexed block at the top level.  We do want to allow that, so
            # we'll pretend it's a block.
            alist = self.classes[Block][0]
            ff = self.classes[Block][1]
        return (alist, ff)

    def get_data_class_attr_list(self, o):
        """
        Look up what attributes to save/load for an ComponentData object.
        Args:
            o: Object to look up attribute list for.
        Return:
            A list of attributes and a filter function for object type
        """
        alist = None  # Attributes to store
        ff = None  # Load filter function
        for cl in self.data_classes:
            if isinstance(o, cl):
                alist = self.data_classes[cl][0]
                ff = self.data_classes[cl][1]
                break
        return (alist, ff)

    @classmethod
    def bound(cls):
        """Returns a StoreSpec object to store variable bounds only."""
        return cls(
            classes={Var: ((), None)},
            data_classes={
                Var._ComponentDataClass: (("lb", "ub"), None),
            },
        )

    @classmethod
    def value(cls, only_not_fixed=False):
        """Returns a StoreSpec object to store variable values only."""
        if only_not_fixed:
            return cls(
                classes={Var: ((), None), BooleanVar: ((), None)},
                data_classes={
                    Var._ComponentDataClass: (("value",), _value_if_not_fixed),
                    BooleanVar._ComponentDataClass: (("value",), _value_if_not_fixed),
                },
            )
        return cls(
            classes={Var: ((), None), BooleanVar: ((), None)},
            data_classes={
                Var._ComponentDataClass: (("value",), None),
                BooleanVar._ComponentDataClass: (("value",), None),
            },
        )

    @classmethod
    def isfixed(cls):
        """Returns a StoreSpec object to store if variables are fixed."""
        return cls(
            classes={Var: ((), None), BooleanVar: ((), None)},
            data_classes={
                Var._ComponentDataClass: (("fixed",), None),
                BooleanVar._ComponentDataClass: (("fixed",), None),
            },
        )

    @classmethod
    def suffix(cls, suffix_filter=None):
        # need to include classes and data that can have suffixes, but don't
        # need to store any attributes
        return cls(
            classes={
                Var: ((), None),
                BooleanVar: ((), None),
                Constraint: ((), None),
                Expression: ((), None),
                Suffix: ((), None),
            },
            data_classes={
                Var._ComponentDataClass: ((), None),
                BooleanVar._ComponentDataClass: ((), None),
                Constraint._ComponentDataClass: ((), None),
                Expression._ComponentDataClass: ((), None),
            },
            suffix_filter=suffix_filter,
        )

    @classmethod
    def value_isfixed(cls, only_fixed):
        """
        Return a StoreSpec object to store variable values and if fixed.

        Args:
            only_fixed: Only load fixed variable values
        """
        if only_fixed:
            return cls(
                classes={Var: ((), None), BooleanVar: ((), None)},
                data_classes={
                    Var._ComponentDataClass: (("value", "fixed"), _only_fixed),
                    BooleanVar._ComponentDataClass: (("value", "fixed"), _only_fixed),
                },
            )
        else:
            return cls(
                classes={Var: ((), None), BooleanVar: ((), None)},
                data_classes={
                    Var._ComponentDataClass: (("value", "fixed"), None),
                    BooleanVar._ComponentDataClass: (("value", "fixed"), None),
                },
            )

    @classmethod
    def value_isfixed_isactive(cls, only_fixed):
        """
        Return a StoreSpec object to store variable values, if variables are
        fixed and if components are active.

        Args:
            only_fixed: Only load fixed variable values
        """
        if only_fixed:
            return cls(
                classes={
                    Var: ((), None),
                    BooleanVar: ((), None),
                    Param: ((), None),
                    Constraint: (("active",), None),
                    Block: (("active",), None),
                },
                data_classes={
                    Var._ComponentDataClass: (("value", "fixed"), _only_fixed),
                    BooleanVar._ComponentDataClass: (("value", "fixed"), _only_fixed),
                    ParamData: (("value",), None),
                    Constraint._ComponentDataClass: (("active",), None),
                    Block._ComponentDataClass: (("active",), None),
                },
            )
        else:
            return cls(
                classes={
                    Var: ((), None),
                    BooleanVar: ((), None),
                    Param: ((), None),
                    Constraint: (("active",), None),
                    Block: (("active",), None),
                },
                data_classes={
                    Var._ComponentDataClass: (("value", "fixed"), None),
                    BooleanVar._ComponentDataClass: (("value", "fixed"), None),
                    ParamData: (("value",), None),
                    Constraint._ComponentDataClass: (("active",), None),
                    Block._ComponentDataClass: (("active",), None),
                },
            )


def _may_have_subcomponents(o):
    """
    Args:
        o: an object.
    Returns:
        True if the object has a callable component_objects method, otherwise
        False.
    """
    if hasattr(o, "component_objects"):
        if hasattr(o.component_objects, "__call__"):
            return True


def _write_component(sd, o, wts, count=None, lookup=None, suffixes=None):
    """
    Writes a component state to the save dictionary under a key given by the
    components name.

    Args:
        sd: dictionary to to save the object into, will create a key that is the
            object name (not fully qualified)
        o: object to save
        wts: a StoreSpec object indicating what object attributes to write
        count: count the number of Pyomo components written also used for ids
        lookup: is a lookup table for component ids from components
        suffixes: is a list of suffixes, that we are delaying writing
    Returns:
        None
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = []

    # Get list of attributes to save, also returns ff, which is a filter
    # function and only used in reading stuff back in.
    alist, ff = wts.get_class_attr_list(o)  # pylint: disable=unused-variable
    if alist is None:
        return  # alist is none means skip this component type
    # Get the component name, doesn't need to be fully qualified or unique because
    # we are storing the state in a hierarchy structure
    oname = o.getname(fully_qualified=False)
    # Create a dictionary for this component, if storing suffixes assign it
    # a sequential id number and create a lookup table that takes the component
    # and returns its id for use later in writing suffix data
    sd[oname] = {"__type__": str(type(o))}
    if Suffix in wts.classes:
        sd[oname]["__id__"] = count.count
    lookup[id(o)] = count.count  # used python id() here for efficiency
    if count is not None:
        count.count += 1  # increment the component counter
    for a in alist:  # store the desired attributes
        if a in wts.write_cbs:
            if wts.write_cbs[a] is None:
                sd[oname][a] = getattr(o, a, None)
            else:
                sd[oname][a] = wts.write_cbs[a](o)
        else:
            sd[oname][a] = getattr(o, a, None)
    sd[oname]["data"] = {}  # create a dict for component data and sub-components
    if isinstance(o, Suffix):  # if is a suffix, make a list and delay writing
        if wts.suffix_filter is None or oname in wts.suffix_filter:
            suffixes.append(
                {"sd": sd[oname]["data"], "o": o, "wts": wts, "lookup": lookup}
            )
    else:  # if not suffix go ahead and write the data
        _write_component_data(
            sd=sd[oname]["data"],
            o=o,
            wts=wts,
            lookup=lookup,
            count=count,
            suffixes=suffixes,
        )


def _write_component_data(sd, o, wts, count=None, lookup=None, suffixes=None):
    """
    Iterate through the component data and write to the sd dictionary. The keys
    for the data items are added to the dictionary. If the component has
    sub-components they are written by a recursive call to _write_component under
    the __pyomo_components__ key.

    Args:
        sd: dictionary to to save the object into, will create keys that are the
            data object indexes repn.
        o: object to save
        wts: a StoreSpec object indicating what object attributes to write
        count: count the number of Pyomo components written also used for ids
        lookup: is a lookup table for component ids from components
        suffixes: is a list of suffixes, that we are delaying writing
    Returns:
        None
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = []

    if isinstance(o, Suffix) and Suffix in wts.classes:
        # make special provision for writing suffixes.
        for key in o:
            el = o[key]
            if id(key) not in lookup:
                # didn't store these components so can't write suffix.
                continue
            sd[lookup[id(key)]] = el  # Assume keys are Pyomo model components
    else:  # rest of components with normal component data structure
        is_first_item = True  # on first item when true
        try:
            item_keys = o.keys()
        except AttributeError:
            item_keys = [None]
        for key in item_keys:
            if key is None and isinstance(o, ComponentData):
                el = o
            else:
                el = o[key]
            if is_first_item:  # assume all item are same type, use first to get alist
                # Get all attributes
                (alist, _) = wts.get_data_class_attr_list(el)
                if alist is None:
                    return  # if None then skip writing
            is_first_item = False  # done with first only stuff
            edict = {"__type__": str(type(el))}
            if Suffix in wts.classes:  # if writing suffixes give data components an id
                edict["__id__"] = count.count
                lookup[id(el)] = count.count  # and add to lookup table
            if count is not None:
                count.count += 1  # increment component counter
            sd[repr(key)] = edict  # stick item dict into component data dict
            for a in alist:  # store desired attributes
                if a in wts.write_cbs:
                    if wts.write_cbs[a] is None:
                        edict[a] = getattr(el, a)
                    else:
                        edict[a] = wts.write_cbs[a](el)
                else:
                    edict[a] = getattr(el, a)
            hascomps = False  # Has sub-components (like a Block would have)
            if _may_have_subcomponents(el):  # block or block like component
                for o2 in el.component_objects(descend_into=False):
                    # loop through sub-components
                    if not hascomps:  # if here it does have sub-components
                        cdict = {}  # so store those in __pyomo_components__
                        edict["__pyomo_components__"] = cdict
                    hascomps = True  # only make __pyomo_components__ dict once
                    # write each sub-component
                    _write_component(
                        sd=cdict,
                        o=o2,
                        wts=wts,
                        count=count,
                        lookup=lookup,
                        suffixes=suffixes,
                    )


def component_data_to_dict(o, wts):
    """
    Component data to a dict.
    """
    el = o
    # get attributes
    alist, ff = wts.get_data_class_attr_list(el)  # pylint: disable=unused-variable
    if alist is None:
        return  # if None then skip writing
    edict = {}  # if not writing suffixes don't need ids
    for a in alist:  # store desired attributes
        edict[a] = getattr(el, a)
    hascomps = False  # Has sub-components (like a Block would have)
    if _may_have_subcomponents(el):  # block or block like component
        for o2 in el.component_objects(descend_into=False):
            # loop through sub-components
            if not hascomps:  # if here it does have sub-components
                cdict = {}  # so store those in __pyomo_components__
                edict["__pyomo_components__"] = cdict
            hascomps = True  # only make __pyomo_components__ dict once
            # write each sub-component
            _write_component(sd=cdict, o=o2, wts=wts)
    return edict


def to_json(
    o,
    fname=None,
    human_read=False,
    wts=None,
    metadata=None,
    gz=None,
    return_dict=False,
    return_json_string=False,
):
    """
    Save the state of a model to a Python dictionary, and optionally dump it
    to a json file.  To load a model state, a model with the same structure must
    exist.  The model itself cannot be recreated from this.

    Args:
        o: The Pyomo component object to save.  Usually a Pyomo model, but could
            also be a sub-component of a model (usually a sub-block).
        fname: json file name to save model state, if None only create
            python dict
        gz: If fname is given and gv is True gzip the json file. The default is
            True if the file name ends with '.gz' otherwise False.
        human_read: if True, add indents and spacing to make the json file more
            readable, if false cut out whitespace and make as compact as
            possible
        metadata: A dictionary of additional metadata to add.
        wts: is What To Save, this is a StoreSpec object that specifies what
            object types and attributes to save.  If None, the default is used
            which saves the state of the complete model state.
        metadata: additional metadata to save beyond the standard format_version,
            date, and time.
        return_dict: default is False if true returns a dictionary representation
        return_json_string: default is False returns a json string

    Returns:
        If return_dict is True returns a dictionary serialization of the Pyomo
        component.  If return_dict is False and return_json_string is True
        returns a json string dump of the dict.  If fname is given the dictionary
        is also written to a json file.  If gz is True and fname is given, writes
        a gzipped json file.
    """
    if gz is None:
        if isinstance(fname, str):
            gz = fname.endswith(".gz")
        else:
            gz = False
    if metadata is None:
        metadata = {}

    suffixes = []
    lookup = {}
    count = Counter()
    start_time = time.time()
    if wts is None:
        wts = StoreSpec()
    now = datetime.datetime.now()
    sd = {
        "__metadata__": {
            "format_version": __format_version__,
            "date": datetime.date.isoformat(now.date()),
            "time": datetime.time.isoformat(now.time()),
            "other": metadata,
        }
    }
    # first write the component
    _write_component(sd, o, wts, count, suffixes=suffixes, lookup=lookup)
    for s in suffixes:
        _write_component_data(**s)
    pdict = {}
    sd["__metadata__"]["__performance__"] = pdict
    pdict["n_components"] = count.count
    dict_time = time.time()
    pdict["etime_make_dict"] = dict_time - start_time
    # This returns the dict but if fname is specified also save to json file
    dump_kw = {"indent": 2} if human_read else {"separators": (",", ":")}
    if fname is not None:
        if gz:
            with gzip.open(fname, "w") as f:
                f.write(json.dumps(sd, **dump_kw).encode("utf-8"))
        else:
            with open(fname, "w") as f:
                json.dump(sd, f, **dump_kw)
    file_time = time.time()
    # unfortunately I can't write how long it took to write the file in the file
    pdict["etime_write_file"] = file_time - dict_time
    if return_dict:
        # In interactive environments returning the dict can cause it to print
        # an extremely large amount of stuff.  So added this option to make sure
        # it's really what you want.
        return sd
    elif return_json_string:
        return json.dumps(sd, **dump_kw)
    else:
        return None


def _read_component(sd, o, wts, lookup=None, suffixes=None, root_name=None):
    """
    Read a component dictionary into a model
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = {}

    alist, ff = wts.get_class_attr_list(o)
    if alist is None:
        return
    if root_name is None:
        oname = o.getname(fully_qualified=False)
    else:
        oname = root_name
    try:
        odict = sd[oname]
    except KeyError as e:
        if wts.ignore_missing:
            return
        else:
            raise e
    if ff is not None:
        alist = ff(o, odict)
    if Suffix in wts.classes:
        lookup[odict["__id__"]] = o
    for a in alist:
        try:
            if a in wts.read_cbs:
                if wts.read_cbs[a] is None:
                    pass
                else:
                    wts.read_cbs[a](o, odict[a])
            else:
                setattr(o, a, odict[a])
        except KeyError as e:
            if wts.ignore_missing:
                return
            else:
                raise e
    if isinstance(o, Suffix):
        if wts.suffix_filter is None or oname in wts.suffix_filter:
            suffixes[odict["__id__"]] = odict["data"]  # is populated
    else:  # read non-suffix component data
        _read_component_data(odict["data"], o, wts, lookup=lookup, suffixes=suffixes)


def _read_component_data(sd, o, wts, lookup=None, suffixes=None):
    """
    Read a Pyomo component's data in from a dict.

    Args:
        sd: dictionary to read from
        o: Pyomo component whose data to read
        wts: StoreSpec object specifying what to read in
        lookup: a lookup table for id to component for reading suffixes
        suffixes: a list of suffixes put off reading until end

    Returns:
        None
    """
    if lookup is None:
        lookup = {}
    if suffixes is None:
        suffixes = {}

    alist = []  # list of attributes to read
    c = 0  # counter of data items in component
    try:
        item_keys = o.keys()
    except AttributeError:
        item_keys = [None]
    for key in item_keys:
        if key is None and isinstance(o, ComponentData):
            el = o
        else:
            el = o[key]
        if c == 0:  # if first data item assume all times are same and get alist
            alist, ff = wts.get_data_class_attr_list(el)  # ff is filter function
            if alist is None:
                return  # skip reading this type
        c += 1
        try:  # get data from dict
            edict = sd[repr(key)]
        except KeyError as e:  # data was missing either ignore or raise except
            if wts.ignore_missing:
                return  # if ignore missing option its okay
            else:
                raise e  # else raise exception
        if ff is not None:  # if a filer function was given, use it to make a
            # new a list based on the model and whats stored for the state
            # this lets you conditionally load things, for example only load
            # values for unfixed variables.
            alist = ff(el, edict)
        if Suffix in wts.classes:  # if loading suffixes make lookup table id
            lookup[edict["__id__"]] = el
        for a in alist:  # read in desired attributes
            try:
                if a in wts.read_cbs:
                    if wts.read_cbs[a] is None:
                        pass
                    else:
                        wts.read_cbs[a](el, edict[a])
                else:  # directly set an attribute
                    setattr(el, a, edict[a])
            except KeyError as e:  # attribute missing
                if wts.ignore_missing:
                    return  # if ignore option then is okay
                else:
                    raise e  # otherwise raise an exception
        if _may_have_subcomponents(el) and "__pyomo_components__" in edict:
            # read sub-components of block-like
            for o2 in el.component_objects(descend_into=False):
                # recursive read here
                _read_component(
                    edict["__pyomo_components__"],
                    o2,
                    wts,
                    lookup=lookup,
                    suffixes=suffixes,
                )


def component_data_from_dict(sd, o, wts):
    """
    Component data to a dict.
    """
    el = o
    alist = []  # list of attributes to read
    alist, ff = wts.get_data_class_attr_list(el)  # ff is filter function
    if alist is None:
        return
    edict = sd
    if ff is not None:
        alist = ff(o, edict)
    for a in alist:  # read in desired attributes
        try:
            if a in wts.read_cbs:  # use a callback
                wts.read_cbs[a](el, edict[a])
            else:  # directly set an attribute
                setattr(el, a, edict[a])
        except KeyError as e:  # attribute missing
            if wts.ignore_missing:
                return  # if ignore option then is okay
            else:
                raise e  # otherwise raise an exception
    if _may_have_subcomponents(el):  # read sub-components of block-like
        for o2 in el.component_objects(descend_into=False):
            # recursive read here
            _read_component(edict["__pyomo_components__"], o2, wts)


def _read_suffixes(lookup, suffixes):
    """
    Go through the list of suffixes and read the data back in.

    Args:
        lookup: a lookup table to go from id to component
        suffixes: a dictionary with suffix id keys and value dict value
    Returns:
        None
    """
    for uid in suffixes:
        d = suffixes[uid]
        s = lookup[uid]  # suffixes keys are ids, so get suffix component
        for key in d:  # set values from value dict
            try:
                kc = lookup[int(key)]  # use int because json turn keys to string
            except KeyError:
                continue
            s[kc] = d[key]


def from_json(o, sd=None, fname=None, s=None, wts=None, gz=None, root_name=None):
    """
    Load the state of a Pyomo component state from a dictionary, json file, or
    json string.  Must only specify one of sd, fname, or s as a non-None value.
    This works by going through the model and loading the state of each
    sub-component of o. If the saved state contains extra information, it is
    ignored.  If the save state doesn't contain an entry for a model component
    that is to be loaded an error will be raised, unless ignore_missing = True.

    Args:
        o: Pyomo component to for which to load state
        sd: State dictionary to load, if None, check fname and s
        fname: JSON file to load, only used if sd is None
        s: JSON string to load only used if both sd and fname are None
        wts: StoreSpec object specifying what to load
        gz: If True assume the file specified by fname is gzipped. The default is
            True if fname ends with '.gz' otherwise False.

    Returns:
        Dictionary with some performance information. The keys are
        "etime_load_file", how long in seconds it took to load the json file
        "etime_read_dict", how long in seconds it took to read models state
        "etime_read_suffixes", how long in seconds it took to read suffixes
    """
    if gz is None:
        if isinstance(fname, str):
            gz = fname.endswith(".gz")
        else:
            gz = False

    # keeping track of elapsed time.  want to make sure I don't do anything
    # that's too slow.
    start_time = time.time()
    # Get the model state dict from one of three sources
    if sd is not None:  # Existing Python dict (for in-memory stuff).
        pass
    elif fname is not None:  # Read in from a json file
        if gz:
            with gzip.open(fname, "r") as f:
                fr = f.read()
                sd = json.loads(fr)
        else:
            with open(fname, "r") as f:
                sd = json.load(f)  # json file
    elif s is not None:  # Use a json string (not really sure if useful)
        sd = json.loads(s)  # json string
    else:  # Didn't specify at least one source
        # PYLINT-TODO
        # pylint: disable-next=broad-exception-raised
        raise Exception("Need to specify a data source to load from")
    dict_time = time.time()  # To calculate how long it took to read file
    if wts is None:  # if no StoreSpec object given use the default, which should
        wts = StoreSpec()  # be the typical save everything important
    lookup = {}  # A dict to use for a lookup tables
    suffixes = {}  # A list of suffixes delayed to end so lookup is complete
    # Read toplevel component (is recursive)
    if root_name is None:
        for k in sd:
            if k.startswith("__") and k.endswith("__"):
                # This is metadata or maybe some similar future addition.
                continue
            else:
                root_name = k
                break  # should be one root, use it's name
    _read_component(sd, o, wts, lookup=lookup, suffixes=suffixes, root_name=root_name)
    read_time = time.time()  # to calc time to read model state minus suffixes
    # Now read in the suffixes
    _read_suffixes(lookup, suffixes)
    suffix_time = time.time()  # to calculate time to read suffixes
    pdict = {}  # return some performance information, to make sure not too slow
    pdict["etime_load_file"] = dict_time - start_time
    pdict["etime_read_dict"] = read_time - dict_time
    pdict["etime_read_suffixes"] = suffix_time - read_time
    return pdict



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""
This is an example of a subcritical boiler recirculation system. This model
does not represent any specific power plant, it shows a generic recirculation
system. This model is for demonstration and tutorial purposes only.

This example consist of a simulation of the internal recirculation system
of a subcritical coal fired boiler. The simulation inlet includes the
feed water flowrate and enthalpy, the unit dimensions/specifications,
and boiler load, which in this case is represented as fixed heat flux from
fire side.

The main outputs of the model are the feedwater pressure (to control the boiler
feedwater pump), and steam outlet to superheaters.

For a detailed description see Jupyter Notebook
authors: Boiler Subsystem Team (J. Ma, M. Zamarripa)
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

import os

# import plotting libraries
import matplotlib.pyplot as plt

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.network import Arc

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util import model_serializer as ms
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

# Import Unit Model Modules
from idaes.models.properties import iapws95

# Import IDAES standard unit model
import idaes.logger as idaeslog
from idaes.models_extra.power_generation.unit_models.drum import Drum
from idaes.models_extra.power_generation.unit_models.downcomer import Downcomer
from idaes.models_extra.power_generation.unit_models.waterwall_section import (
    WaterwallSection,
)
from idaes.models_extra.power_generation.properties.flue_gas_ideal import (
    FlueGasParameterBlock,
)

_log = idaeslog.getModelLogger(__name__)


def main(m=None):
    """Create concrete model, make the flowsheet object, fix some variables,
    and solve the problem.
    if m concrete model already exists this function only builds the unit
    models, fixes the inputs, and initializes the units.
    This is useful to import subsections of a flowsheet
    (i.e. boiler subsystem + steam cycle) if concrete model already exists
    we assume m.fs flowsheet, m.fs.prop_water and m.fs.prop_gas also exist"""
    if m is None:
        # Create a Concrete Model as the top level object
        m = pyo.ConcreteModel()
        # Add a flowsheet object to the model
        m.fs = FlowsheetBlock(dynamic=False)
        # Add property packages to flowsheet library
        m.fs.prop_water = iapws95.Iapws95ParameterBlock()
        m.fs.prop_gas = FlueGasParameterBlock()

    create_model(m)
    set_inputs(m)
    initialize(m)
    return m


def create_model(m):
    """Create unit models"""

    m.fs.ww_zones = pyo.RangeSet(10)

    m.fs.drum = Drum(
        property_package=m.fs.prop_water,
        has_holdup=False,
        has_heat_transfer=True,
        has_pressure_change=True,
    )
    m.fs.downcomer = Downcomer(
        property_package=m.fs.prop_water, has_holdup=False, has_heat_transfer=True
    )
    m.fs.Waterwalls = WaterwallSection(
        m.fs.ww_zones,
        dynamic=False,
        has_holdup=False,
        property_package=m.fs.prop_water,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    m.fs.stream_drum_out = Arc(
        source=m.fs.drum.liquid_outlet, destination=m.fs.downcomer.inlet
    )
    m.fs.stream_dcmr_out = Arc(
        source=m.fs.downcomer.outlet, destination=m.fs.Waterwalls[1].inlet
    )

    def arc_rule(b, i):
        return {
            "source": m.fs.Waterwalls[i].outlet,
            "destination": m.fs.Waterwalls[i + 1].inlet,
        }

    m.arc = Arc(pyo.RangeSet(9), rule=arc_rule)

    m.fs.stream_ww14 = Arc(
        source=m.fs.Waterwalls[10].outlet, destination=m.fs.drum.water_steam_inlet
    )
    # pyomo arcs expand constraints for inlet = outlet ports
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)


def set_inputs(m):
    """fix variables for geometry and design data"""

    # drum
    m.fs.drum.drum_diameter.fix(1.2)
    m.fs.drum.drum_length.fix(15.3256)
    m.fs.drum.number_downcomers.fix(6)
    m.fs.drum.downcomer_diameter.fix(0.38)
    m.fs.drum.drum_level[:].fix(0.6)  # drum level, set as unit radius
    m.fs.drum.heat_duty[:].fix(0.0)  # assume no heat loss
    # downcomer inputs
    m.fs.downcomer.diameter.fix(0.3)
    m.fs.downcomer.number_downcomers.fix(6)  # number of downcomers
    m.fs.downcomer.height.fix(40.1)
    m.fs.downcomer.heat_duty[:].fix(0.0)

    # 10 waterwall sections
    for i in m.fs.ww_zones:
        m.fs.Waterwalls[i].tube_diameter.fix(0.047)
        m.fs.Waterwalls[i].tube_thickness.fix(0.00350)
        m.fs.Waterwalls[i].fin_thickness.fix(0.00455)
        m.fs.Waterwalls[i].slag_thickness[:].fix(0.001)
        m.fs.Waterwalls[i].fin_length.fix(0.0115)
        m.fs.Waterwalls[i].number_tubes.fix(610)
        m.fs.Waterwalls[i].fcorrection_dp.fix(1.2)

    # water wall section height (must be same as fire side model zones)
    m.fs.Waterwalls[1].height.fix(6.150)
    m.fs.Waterwalls[2].height.fix(3.150)
    m.fs.Waterwalls[3].height.fix(1.5)
    m.fs.Waterwalls[4].height.fix(1.450)
    m.fs.Waterwalls[5].height.fix(1.350)
    m.fs.Waterwalls[6].height.fix(1.250)
    m.fs.Waterwalls[7].height.fix(1.150)
    m.fs.Waterwalls[8].height.fix(1.350)
    m.fs.Waterwalls[9].height.fix(3.250)
    m.fs.Waterwalls[10].height.fix(3.450)

    # water wall section projected area
    m.fs.Waterwalls[1].projected_area.fix(320.0)
    m.fs.Waterwalls[2].projected_area.fix(150.3)
    m.fs.Waterwalls[3].projected_area.fix(70.8)
    m.fs.Waterwalls[4].projected_area.fix(70.0)
    m.fs.Waterwalls[5].projected_area.fix(58.6)
    m.fs.Waterwalls[6].projected_area.fix(58.6)
    m.fs.Waterwalls[7].projected_area.fix(50.1)
    m.fs.Waterwalls[8].projected_area.fix(65.6)
    m.fs.Waterwalls[9].projected_area.fix(145.6)
    m.fs.Waterwalls[10].projected_area.fix(165.5)


def initialize(
    m,
    outlvl=idaeslog.NOTSET,
    optarg=None,
):
    """Initialize unit models"""
    init_log = idaeslog.getInitLogger(m.name, outlvl, tag="flowsheet")
    solve_log = idaeslog.getSolveLogger(m.name, outlvl, tag="flowsheet")

    if optarg is None:
        optarg = {
            "tol": 1e-6,
            "max_iter": 40,
        }

    solver = get_solver()
    solver.options = optarg
    init_log.info_low("Starting initialization...")

    if not os.path.exists("subcritical_boiler_init.json.gz"):
        # 10 Waterwalls, initial guess for specific simulation
        m.fs.Waterwalls[1].heat_fireside[:].fix(2.3e7)
        m.fs.Waterwalls[2].heat_fireside[:].fix(1.5e7)
        m.fs.Waterwalls[3].heat_fireside[:].fix(6.8e6)
        m.fs.Waterwalls[4].heat_fireside[:].fix(1.2e7)
        m.fs.Waterwalls[5].heat_fireside[:].fix(1.2e7)
        m.fs.Waterwalls[6].heat_fireside[:].fix(1.2e7)
        m.fs.Waterwalls[7].heat_fireside[:].fix(1.0e7)
        m.fs.Waterwalls[8].heat_fireside[:].fix(9.9e6)
        m.fs.Waterwalls[9].heat_fireside[:].fix(2.1)
        m.fs.Waterwalls[10].heat_fireside[:].fix(2.0e7)

        state_args_water_steam = {
            "flow_mol": 199470.7831,  # mol/s
            "pressure": 10903981.9107,  # Pa
            "enth_mol": 26585.3483,
        }  # j/mol

        state_args_feedwater = {
            "flow_mol": 4630.6098,  # mol/s
            "pressure": 10903981.9107,  # Pa
            "enth_mol": 22723.907,
        }  # j/mol

        m.fs.drum.initialize(
            outlvl=outlvl,
            optarg=solver.options,
            state_args_water_steam=state_args_water_steam,
            state_args_feedwater=state_args_feedwater,
        )
        m.fs.downcomer.inlet.flow_mol[:].fix(m.fs.drum.liquid_outlet.flow_mol[0].value)
        m.fs.downcomer.inlet.pressure[:].fix(m.fs.drum.liquid_outlet.pressure[0].value)
        m.fs.downcomer.inlet.enth_mol[:].fix(m.fs.drum.liquid_outlet.enth_mol[0].value)

        m.fs.downcomer.initialize(
            state_args={
                "flow_mol": m.fs.drum.liquid_outlet.flow_mol[0].value,
                "pressure": m.fs.drum.liquid_outlet.pressure[0].value,
                "enth_mol": m.fs.drum.liquid_outlet.enth_mol[0].value,
            },
            outlvl=outlvl,
            optarg=solver.options,
        )

        m.fs.Waterwalls[1].inlet.flow_mol[:].fix(
            m.fs.downcomer.outlet.flow_mol[0].value
        )
        m.fs.Waterwalls[1].inlet.pressure[:].fix(
            m.fs.downcomer.outlet.pressure[0].value
        )
        m.fs.Waterwalls[1].inlet.enth_mol[:].fix(
            m.fs.downcomer.outlet.enth_mol[0].value
        )
        m.fs.Waterwalls[1].initialize(
            state_args={
                "flow_mol": m.fs.Waterwalls[1].inlet.flow_mol[0].value,
                "pressure": m.fs.Waterwalls[1].inlet.pressure[0].value,
                "enth_mol": m.fs.Waterwalls[1].inlet.enth_mol[0].value,
            },
            outlvl=6,
            optarg=solver.options,
        )

        for i in range(2, 11):
            m.fs.Waterwalls[i].inlet.flow_mol[:].fix(
                m.fs.Waterwalls[i - 1].outlet.flow_mol[0].value
            )
            m.fs.Waterwalls[i].inlet.pressure[:].fix(
                m.fs.Waterwalls[i - 1].outlet.pressure[0].value
            )
            m.fs.Waterwalls[i].inlet.enth_mol[:].fix(
                m.fs.Waterwalls[i - 1].outlet.enth_mol[0].value
            )
            m.fs.Waterwalls[i].initialize(
                state_args={
                    "flow_mol": m.fs.Waterwalls[i - 1].outlet.flow_mol[0].value,
                    "pressure": m.fs.Waterwalls[i - 1].outlet.pressure[0].value,
                    "enth_mol": m.fs.Waterwalls[i - 1].outlet.enth_mol[0].value,
                },
                outlvl=6,
                optarg=solver.options,
            )

        m.fs.drum.feedwater_inlet.flow_mol[:].fix()
        m.fs.drum.feedwater_inlet.pressure[:].unfix()
        m.fs.drum.feedwater_inlet.enth_mol[:].fix()

        m.fs.downcomer.inlet.flow_mol[:].unfix()
        m.fs.downcomer.inlet.pressure[:].unfix()
        m.fs.downcomer.inlet.enth_mol[:].unfix()
        print("solving full-space problem")
        for i in m.fs.ww_zones:
            m.fs.Waterwalls[i].inlet.flow_mol[:].unfix()
            m.fs.Waterwalls[i].inlet.pressure[:].unfix()
            m.fs.Waterwalls[i].inlet.enth_mol[:].unfix()

        df = degrees_of_freedom(m)
        if df != 0:
            raise ValueError("Check degrees of freedom: {}".format(df))
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            print("solving full-space problem")
            res = solver.solve(m, tee=slc.tee)
        init_log.info_low("Initialization Complete: {}".format(idaeslog.condition(res)))
        ms.to_json(m, fname="subcritical_boiler_init.json.gz")
    else:
        print("\n\nInitializing from json file")
        ms.from_json(m, fname="subcritical_boiler_init.json.gz")
    return m


def run_sensitivity():
    m = main()
    optarg = {"tol": 1e-6, "max_iter": 40}

    m.fs.drum.feedwater_inlet.flow_mol[:].fix()
    m.fs.drum.feedwater_inlet.pressure[:].unfix()
    m.fs.drum.feedwater_inlet.enth_mol[:].fix()
    # set scaling parameters
    for i in m.fs.ww_zones:
        iscale.set_scaling_factor(m.fs.Waterwalls[i].heat_flux_conv[0], 1e-5)
    iscale.calculate_scaling_factors(m)
    # solve flowsheet
    solver = get_solver()
    solver.options = optarg
    results = solver.solve(m, tee=True)
    print(results)

    print("\nDrum feedwater inlet")
    m.fs.drum.feedwater_inlet.display()
    print("\nWaterWall Outlet")
    m.fs.drum.water_steam_inlet.display()
    m.fs.Waterwalls[10].outlet.display()
    print("\ndrum flash inlet")
    m.fs.drum.flash.inlet.display()
    print("\ndrum liquid outlet")
    m.fs.drum.liquid_outlet.display()
    print("\ndowncomer inlet")
    m.fs.downcomer.inlet.display()
    print("\ndowncomer outlet")
    m.fs.downcomer.outlet.display()
    print("\nwaterwall inlet")
    m.fs.Waterwalls[1].inlet.display()
    print("\n vapor fraction zone 10")
    m.fs.Waterwalls[10].control_volume.properties_out[0].vapor_frac.display()

    # Since heat transfer indirectly calculates the recirculation flowrate
    # and drum level is fixed in this case, the feedwater flowrate
    # can be calculated as a function of the boiler load, by changing the
    # heat duty to waterwall.
    # This sensitivity analysis demonstrate that increasing the boiler load
    # coal flowrate (in this case heat duty), increases the feedwater flowrate
    # and the steam outlet
    vap_frac = []
    flow = []
    flow_it = [
        1.1,
        1.2,
        1.3,
        1.4,
        1.5,
        1.6,
        1.7,
        1.8,
        1.9,
        2.0,
        2.1,
        2.2,
        2.3,
        2.4,
        2.5,
        2.6,
        2.7,
        2.8,
        2.9,
        3.0,
    ]
    count = 0
    for i in flow_it:
        count = count + 1
        m.fs.Waterwalls[1].heat_fireside[:].fix(2.3e7 * i)
        m.fs.Waterwalls[2].heat_fireside[:].fix(1.5e7 * i)
        m.fs.Waterwalls[3].heat_fireside[:].fix(6.8e6 * i)
        m.fs.Waterwalls[4].heat_fireside[:].fix(1.2e7 * i)
        m.fs.Waterwalls[5].heat_fireside[:].fix(1.2e7 * i)
        m.fs.Waterwalls[6].heat_fireside[:].fix(1.2e7 * i)
        m.fs.Waterwalls[7].heat_fireside[:].fix(1.0e7 * i)
        m.fs.Waterwalls[8].heat_fireside[:].fix(9.9e6 * i)
        m.fs.Waterwalls[9].heat_fireside[:].fix(2.1e7 * i)
        m.fs.Waterwalls[10].heat_fireside[:].fix(2.0e7 * i)
        m.fs.drum.feedwater_inlet.flow_mol[:].unfix()
        m.fs.drum.feedwater_inlet.pressure[:].fix()
        m.fs.drum.feedwater_inlet.enth_mol[:].fix()
        results = solver.solve(m, tee=False)
        vap_frac.append(
            pyo.value(m.fs.Waterwalls[10].control_volume.properties_out[0].vapor_frac)
        )
        flow.append(pyo.value(m.fs.drum.feedwater_inlet.flow_mol[0]))

        if pyo.check_optimal_termination(results):
            print("iter " + str(count) + " = EXIT- Optimal Solution Found.")
        else:
            print("iter " + str(count) + " = infeasible")

    x = []
    for i in flow_it:
        x.append((i - 1) * 100)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot(x, flow, label=str("feedwater (steam) flowrate"))
    plt.legend()
    ax.set_xlabel("heat duty - % increment")
    ax.set_ylabel("Flowrate mol/s")
    plt.show()
    return m



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Unit operation model for a steam heater applicable to platen superheater
and roof superheater, model main equations:

* Heat is given by fire-side boiler model
* Calculate pressure change due to friction
* Calculate slag layer wall temperature
* Consider a layer of metal and a layer of slag

Created on Aug 27, 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.models.properties import iapws95

from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver
from idaes.models_extra.power_generation.unit_models.steamheater import SteamHeater

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_unit():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = SteamHeater(
        dynamic=False,
        property_package=m.fs.properties,
        has_holdup=True,
        has_heat_transfer=True,
        has_pressure_change=True,
        single_side_only=True,
    )
    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a steam heater model and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 12
    # Check unit config arguments
    assert len(m.fs.unit.config) == 10
    assert m.fs.unit.config.has_heat_transfer
    assert isinstance(m.fs.unit.heat_duty, pyo.Var)
    assert isinstance(m.fs.unit.deltaP, pyo.Var)
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_unit):
    m = build_unit
    # platen superheater
    m.fs.unit.diameter_in.fix(0.035)
    m.fs.unit.tube_thickness.fix(0.005)
    m.fs.unit.fin_thickness.fix(0.0035)
    m.fs.unit.slag_thickness[:].fix(0.001)
    m.fs.unit.fin_length.fix(0.00966)
    m.fs.unit.tube_length.fix(40.0)
    m.fs.unit.number_tubes.fix(200.0)
    m.fs.unit.therm_cond_slag.fix(1.3)
    m.fs.unit.control_volume.scaling_factor_holdup_energy = 1e-8
    m.fs.unit.heat_fireside[:].fix(5.5e7)  # initial guess
    hpl = iapws95.htpx(798.15 * pyo.units.K, 24790249.01 * pyo.units.Pa)
    m.fs.unit.inlet[:].flow_mol.fix(24194.177)
    m.fs.unit.inlet[:].enth_mol.fix(hpl)
    m.fs.unit.inlet[:].pressure.fix(24790249.01)

    state_args = {"flow_mol": 24194.177, "pressure": 24790249.01, "enth_mol": hpl}

    initialization_tester(build_unit, dof=0, state_args=state_args)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unit(build_unit):
    m = build_unit

    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    # check material balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].flow_mol
                - m.fs.unit.control_volume.properties_out[0].flow_mol
            ),
            abs=1e-3,
        )
        == 0
    )

    # pressure drop
    assert pytest.approx(-282158.4020, rel=1e-5) == pyo.value(m.fs.unit.deltaP[0])
    # check energy balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].enth_mol
                - m.fs.unit.control_volume.properties_out[0].enth_mol
            ),
            abs=1e-3,
        )
        == -2273.2742
    )

    assert pytest.approx(835.7582, abs=1e-3) == pyo.value(
        m.fs.unit.control_volume.properties_out[0].temperature
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Reboiler model for distillation.

While the reboiler model, is fairly simple, a major
portion of this code has gone into making this generic and be able to handle
different state variables and the associated splits.
"""

__author__ = "Jaffer Ghouse"

from pandas import DataFrame

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.network import Port
from pyomo.environ import (
    check_optimal_termination,
    Reference,
    Var,
    Constraint,
    value,
    Set,
)

# Import IDAES cores
import idaes.logger as idaeslog
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    EnergyBalanceType,
    MomentumBalanceType,
    MaterialBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import (
    PropertyNotSupportedError,
    ConfigurationError,
    InitializationError,
)
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models_extra.column_models.util import make_phase_split


_log = idaeslog.getIdaesLogger(__name__)


@declare_process_block_class("Reboiler")
class ReboilerData(UnitModelBlockData):
    """
    Reboiler unit for distillation model.
    Unit model to reboil the liquid from the bottom tray of
    the distillation column.
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "has_boilup_ratio",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Boilup ratio term construction flag",
            doc="""Indicates whether terms for boilup ratio should be
constructed,
**default** - False.
**Valid values:** {
**True** - include construction of boilup ratio constraint,
**False** - exclude construction of boilup ratio constraint}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """Build the model.
        Args:
            None
        Returns:
            None
        """
        # Setup model build logger
        model_log = idaeslog.getModelLogger(self.name, tag="unit")

        # Call UnitModel.build to setup dynamics
        super(ReboilerData, self).build()

        # Add Control Volume for the Reboiler
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=True)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type, has_phase_equilibrium=True
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_heat_transfer=True
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # Get liquid and vapor phase objects from the property package
        # to be used below. Avoids repetition.
        _liquid_list = []
        _vapor_list = []
        for p in self.control_volume.config.property_package.phase_list:
            pobj = self.control_volume.config.property_package.get_phase(p)
            if pobj.is_vapor_phase():
                _vapor_list.append(p)
            elif pobj.is_liquid_phase():
                _liquid_list.append(p)
            else:
                _liquid_list.append(p)
                model_log.warning(
                    "A non-liquid/non-vapor phase was detected but will "
                    "be treated as a liquid."
                )

        # Create a pyomo set for indexing purposes. This set is appended to
        # model otherwise results in an abstract set.
        self._liquid_set = Set(initialize=_liquid_list)
        self._vapor_set = Set(initialize=_vapor_list)

        if self.config.has_boilup_ratio is True:

            self.boilup_ratio = Var(initialize=0.5, doc="Boilup ratio for reboiler")

            def rule_boilup_ratio(self, t):
                if hasattr(self.control_volume.properties_out[t], "flow_mol_phase"):
                    return self.boilup_ratio * sum(
                        self.control_volume.properties_out[t].flow_mol_phase[p]
                        for p in self._liquid_set
                    ) == sum(
                        self.control_volume.properties_out[t].flow_mol_phase["Vap"]
                        for p in self._vapor_set
                    )
                elif hasattr(
                    self.control_volume.properties_out[t], "flow_mol_phase_comp"
                ):
                    return self.boilup_ratio * sum(
                        self.control_volume.properties_out[t].flow_mol_phase_comp[p, i]
                        for p in self._liquid_set
                        for i in self.control_volume.properties_out[
                            t
                        ].params.component_list
                    ) == sum(
                        self.control_volume.properties_out[t].flow_mol_phase_comp[p, i]
                        for p in self._vapor_set
                        for i in self.control_volume.properties_out[
                            t
                        ].params.component_list
                    )
                else:
                    raise PropertyNotSupportedError(
                        "Unrecognized names for flow variables encountered "
                        "while building the constraint for reboiler."
                    )

            self.eq_boilup_ratio = Constraint(
                self.flowsheet().time, rule=rule_boilup_ratio
            )

        self.add_inlet_port()

        # Outlet ports that always exist irrespective of reboiler type
        self.bottoms = Port(noruleinit=True, doc="Bottoms stream.")

        self.vapor_reboil = Port(
            noruleinit=True,
            doc="Vapor outlet stream that is returned to " "to the bottom tray.",
        )

        make_phase_split(
            self.control_volume,
            port=self.bottoms,
            phase=self._liquid_set,
            side_sf=1,
            equipmentType="Reboiler",
        )

        make_phase_split(
            self.control_volume,
            port=self.vapor_reboil,
            phase=self._vapor_set,
            side_sf=1,
            equipmentType="Reboiler",
        )

        # Add object reference to variables of the control volume
        # Reference to the heat duty
        self.heat_duty = Reference(self.control_volume.heat[:])

        # Reference to the pressure drop (if set to True)
        if self.config.has_pressure_change:
            self.deltaP = Reference(self.control_volume.deltaP[:])

    def _make_ports(self):

        # Add Ports for the reboiler
        # Inlet port (the vapor from the top tray)
        self.add_inlet_port()

        # Outlet ports that always exist irrespective of reboiler type
        self.bottoms = Port(noruleinit=True, doc="Bottoms stream.")

        self.vapor_reboil = Port(
            noruleinit=True,
            doc="Vapor outlet stream that is returned to " "to the bottom tray.",
        )

    def initialize(
        self, state_args=None, solver=None, optarg=None, outlvl=idaeslog.NOTSET
    ):

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        solverobj = get_solver(solver, optarg)

        # Initialize the inlet and outlet state blocks. Calling the state
        # blocks initialize methods directly so that custom set of state args
        # can be passed to the inlet and outlet state blocks as control_volume
        # initialize method initializes the state blocks with the same
        # state conditions.
        flags = self.control_volume.properties_in.initialize(
            state_args=state_args,
            solver=solver,
            optarg=optarg,
            outlvl=outlvl,
            hold_state=True,
        )

        # Initialize outlet state block at same conditions of inlet except
        # the temperature. Set the temperature to a temperature guess based
        # on the desired boilup_ratio.

        # Get index for bubble point temperature and and assume it
        # will have only a single phase equilibrium pair. This is to
        # support the generic property framework where the T_bubble
        # is indexed by the phases_in_equilibrium. In distillation,
        # the assumption is that there will only be a single pair
        # i.e. vap-liq.
        idx = next(iter(self.control_volume.properties_in[0].temperature_bubble))
        temp_guess = 0.5 * (
            value(self.control_volume.properties_in[0].temperature_dew[idx])
            - value(self.control_volume.properties_in[0].temperature_bubble[idx])
        ) + value(self.control_volume.properties_in[0].temperature_bubble[idx])

        state_args_outlet = {}
        state_dict_outlet = self.control_volume.properties_in[
            self.flowsheet().time.first()
        ].define_port_members()

        for k in state_dict_outlet.keys():
            if state_dict_outlet[k].is_indexed():
                state_args_outlet[k] = {}
                for m in state_dict_outlet[k].keys():
                    state_args_outlet[k][m] = value(state_dict_outlet[k][m])
            else:
                if k != "temperature":
                    state_args_outlet[k] = value(state_dict_outlet[k])
                else:
                    state_args_outlet[k] = temp_guess

        self.control_volume.properties_out.initialize(
            state_args=state_args_outlet,
            solver=solver,
            optarg=optarg,
            outlvl=outlvl,
            hold_state=False,
        )

        if degrees_of_freedom(self) == 0:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solverobj.solve(self, tee=slc.tee)
            init_log.info(
                "Initialization Complete, {}.".format(idaeslog.condition(res))
            )
        else:
            raise ConfigurationError(
                "State vars fixed but degrees of freedom "
                "for reboiler is not zero during "
                "initialization. Please ensure that the boilup_ratio "
                "or the outlet temperature is fixed."
            )

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        self.control_volume.properties_in.release_state(flags=flags, outlvl=outlvl)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        stream_attributes = {}

        stream_dict = {
            "Inlet": "inlet",
            "Vapor Reboil": "vapor_reboil",
            "Bottoms": "bottoms",
        }

        for n, v in stream_dict.items():
            port_obj = getattr(self, v)

            stream_attributes[n] = {}

            for k in port_obj.vars:
                for i in port_obj.vars[k].keys():
                    if isinstance(i, float):
                        stream_attributes[n][k] = value(port_obj.vars[k][time_point])
                    else:
                        if len(i) == 2:
                            kname = str(i[1])
                        else:
                            kname = str(i[1:])
                        stream_attributes[n][k + " " + kname] = value(
                            port_obj.vars[k][time_point, i[1:]]
                        )

        return DataFrame.from_dict(stream_attributes, orient="columns")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Integration tests for generic property package framework

Author: Andrew Lee
"""
# Import Python libraries
import pytest

# Import Pyomo components
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Set,
    value,
    units as pyunits,
)

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component, PhaseType as PT
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity

import idaes.models.properties.modular_properties.pure.RPP4 as RPP4
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available

import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
configuration = {
    # Specifying components
    "components": {
        "benzene": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": 78.1136e-3,  # [1]
                "pressure_crit": 48.9e5,  # [1]
                "temperature_crit": 562.2,  # [1]
                "omega": 0.212,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": -3.392e1,  # [1]
                    "B": 4.739e-1,
                    "C": -3.017e-4,
                    "D": 7.130e-8,
                },
                "enth_mol_form_vap_comp_ref": 82.9e3,  # [3]
                "entr_mol_form_vap_comp_ref": -269,  # [3]
                "pressure_sat_comp_coeff": {
                    "A": -6.98273,  # [1]
                    "B": 1.33213,
                    "C": -2.62863,
                    "D": -3.33399,
                },
            },
        },
        "toluene": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": 92.1405e-3,  # [1]
                "pressure_crit": 41e5,  # [1]
                "temperature_crit": 591.8,  # [1]
                "omega": 0.263,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": -2.435e1,
                    "B": 5.125e-1,
                    "C": -2.765e-4,
                    "D": 4.911e-8,
                },
                "enth_mol_form_vap_comp_ref": 50.1e3,  # [3]
                "entr_mol_form_vap_comp_ref": -321,  # [3]
                "pressure_sat_comp_coeff": {
                    "A": -7.28607,  # [1]
                    "B": 1.38091,
                    "C": -2.83433,
                    "D": -2.79168,
                },
            },
        },
        "l_only": {
            "type": Component,
            "valid_phase_types": PT.liquidPhase,
            "enth_mol_ig_comp": RPP4,
            "parameter_data": {
                "mw": 92.1405e-3,
                "pressure_crit": 41e5,
                "temperature_crit": 591.8,
                "omega": 0.263,
                "cp_mol_ig_comp_coeff": {
                    "A": -2.435e1,
                    "B": 5.125e-1,
                    "C": -2.765e-4,
                    "D": 4.911e-8,
                },
                "enth_mol_form_vap_comp_ref": 50.1e3,
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000),
        "temperature": (273.15, 300, 450),
        "pressure": (5e4, 1e5, 1e6),
    },
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    # Declare a base units dict to save code later
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("benzene", "benzene"): 0.000,
            ("benzene", "toluene"): 0.000,
            ("benzene", "l_only"): 0.000,
            ("toluene", "benzene"): 0.000,
            ("toluene", "toluene"): 0.000,
            ("toluene", "l_only"): 0.000,
            ("l_only", "benzene"): 0.000,
            ("l_only", "toluene"): 0.000,
            ("l_only", "l_only"): 0.000,
        }
    },
}


class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 3
        for i in model.params.component_list:
            assert i in ["benzene", "toluene", "l_only"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 5
        for i in model.params._phase_component_set:
            assert i in [
                ("Liq", "benzene"),
                ("Liq", "toluene"),
                ("Vap", "benzene"),
                ("Vap", "toluene"),
                ("Liq", "l_only"),
            ]

        assert model.params.config.state_definition == FTPx

        assert model.params.config.state_bounds == {
            "flow_mol": (0, 100, 1000),
            "temperature": (273.15, 300, 450),
            "pressure": (5e4, 1e5, 1e6),
        }

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 2
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2"]

        assert model.params.phase_equilibrium_list == {
            "PE1": ["benzene", ("Vap", "Liq")],
            "PE2": ["toluene", ("Vap", "Liq")],
        }

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300


class TestNonVapourisable_Vapour(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(380)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["benzene"].fix(0.4)
        model.props[1].mole_frac_comp["toluene"].fix(0.4)
        model.props[1].mole_frac_comp["l_only"].fix(0.2)

        return model

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model.props[1]) == 0

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.component
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6}, outlvl=idaeslog.DEBUG)

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.component
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.component
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "benzene"
        ].value == pytest.approx(0.3066, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "toluene"
        ].value == pytest.approx(0.4119, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "l_only"
        ].value == pytest.approx(0.2815, abs=1e-4)

        assert pytest.approx(1, abs=1e-4) == (
            model.props[1].mole_frac_phase_comp["Liq", "benzene"].value
            + model.props[1].mole_frac_phase_comp["Liq", "toluene"].value
            + model.props[1].mole_frac_phase_comp["Liq", "l_only"].value
        )

        assert model.props[1].mole_frac_phase_comp[
            "Vap", "benzene"
        ].value == pytest.approx(0.6291, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Vap", "toluene"
        ].value == pytest.approx(0.3709, abs=1e-4)

        assert pytest.approx(1, abs=1e-4) == value(
            model.props[1].mole_frac_phase_comp["Vap", "benzene"]
            + model.props[1].mole_frac_phase_comp["Vap", "toluene"]
        )

        assert model.props[1].phase_frac["Vap"].value == pytest.approx(0.2896, abs=1e-4)
        assert model.props[1].phase_frac["Liq"].value == pytest.approx(0.7104, abs=1e-4)

        assert pytest.approx(
            model.props[1].mole_frac_comp["benzene"].value, abs=1e-4
        ) == value(
            model.props[1].mole_frac_phase_comp["Vap", "benzene"]
            * model.props[1].phase_frac["Vap"]
            + model.props[1].mole_frac_phase_comp["Liq", "benzene"]
            * model.props[1].phase_frac["Liq"]
        )
        assert pytest.approx(
            model.props[1].mole_frac_comp["toluene"].value, abs=1e-4
        ) == value(
            model.props[1].mole_frac_phase_comp["Vap", "toluene"]
            * model.props[1].phase_frac["Vap"]
            + model.props[1].mole_frac_phase_comp["Liq", "toluene"]
            * model.props[1].phase_frac["Liq"]
        )
        assert pytest.approx(
            model.props[1].mole_frac_comp["l_only"].value, abs=1e-4
        ) == value(
            model.props[1].mole_frac_phase_comp["Liq", "l_only"]
            * model.props[1].phase_frac["Liq"]
        )

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    @pytest.mark.ui
    def test_report(self, model):
        model.props[1].report()


class TestNonVapourisable_Liquid(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        model.props = model.params.build_state_block([1], defined_state=True)

        # Fix state
        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(360)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["benzene"].fix(0.4)
        model.props[1].mole_frac_comp["toluene"].fix(0.4)
        model.props[1].mole_frac_comp["l_only"].fix(0.2)

        return model

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model.props[1]) == 0

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.component
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6}, outlvl=idaeslog.DEBUG)

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.component
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.component
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "benzene"
        ].value == pytest.approx(0.4, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "toluene"
        ].value == pytest.approx(0.4, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "l_only"
        ].value == pytest.approx(0.2, abs=1e-4)

        assert pytest.approx(1, abs=1e-4) == (
            model.props[1].mole_frac_phase_comp["Liq", "benzene"].value
            + model.props[1].mole_frac_phase_comp["Liq", "toluene"].value
            + model.props[1].mole_frac_phase_comp["Liq", "l_only"].value
        )

        assert model.props[1].mole_frac_phase_comp[
            "Vap", "benzene"
        ].value == pytest.approx(0.6995, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Vap", "toluene"
        ].value == pytest.approx(0.3005, abs=1e-4)

        assert pytest.approx(1, abs=1e-4) == value(
            model.props[1].mole_frac_phase_comp["Vap", "benzene"]
            + model.props[1].mole_frac_phase_comp["Vap", "toluene"]
        )

        assert model.props[1].phase_frac["Vap"].value == pytest.approx(0, abs=1e-4)
        assert model.props[1].phase_frac["Liq"].value == pytest.approx(1, abs=1e-4)

        assert pytest.approx(
            model.props[1].mole_frac_comp["benzene"].value, abs=1e-4
        ) == value(
            model.props[1].mole_frac_phase_comp["Vap", "benzene"]
            * model.props[1].phase_frac["Vap"]
            + model.props[1].mole_frac_phase_comp["Liq", "benzene"]
            * model.props[1].phase_frac["Liq"]
        )
        assert pytest.approx(
            model.props[1].mole_frac_comp["toluene"].value, abs=1e-4
        ) == value(
            model.props[1].mole_frac_phase_comp["Vap", "toluene"]
            * model.props[1].phase_frac["Vap"]
            + model.props[1].mole_frac_phase_comp["Liq", "toluene"]
            * model.props[1].phase_frac["Liq"]
        )
        assert pytest.approx(
            model.props[1].mole_frac_comp["l_only"].value, abs=1e-4
        ) == value(
            model.props[1].mole_frac_phase_comp["Liq", "l_only"]
            * model.props[1].phase_frac["Liq"]
        )

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    @pytest.mark.ui
    def test_report(self, model):
        model.props[1].report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Test the parts of component registry that aren't usually hit."""

import pytest

from idaes.models.properties.general_helmholtz.components import (
    register_helmholtz_component,
    remove_component,
    registered_components,
    component_registered,
    viscosity_available,
    thermal_conductivity_available,
    surface_tension_available,
    component_registered,
    eos_reference,
    viscosity_reference,
    thermal_conductivity_reference,
    surface_tension_reference,
)


@pytest.mark.unit
def test_not_registered():
    assert not viscosity_available("not real")
    assert not thermal_conductivity_available("not real")
    assert not surface_tension_available("not real")
    assert not component_registered("not real")


@pytest.mark.unit
def test_remove_not_registered():
    with pytest.raises(KeyError):
        remove_component("not real")


@pytest.mark.unit
def test_remove_registered():
    register_helmholtz_component("fooponent")
    assert component_registered("fooponent")
    remove_component("fooponent")
    assert not component_registered("fooponent")


@pytest.mark.unit
def test_registered_components():
    """Use some standard components to test list"""
    assert "h2o" in registered_components()
    assert "co2" in registered_components()
    assert "r134a" in registered_components()


@pytest.mark.unit
def test_register_and_get():
    register_helmholtz_component(
        "fooponent",
        viscosity=True,
        thermal_conductivity=True,
        surface_tension=True,
        eos_ref="cite eos",
        viscosity_ref="cite viscosity",
        thermal_conductivity_ref="cite thermal conductivity",
        surface_tension_ref="cite surface tension",
    )
    assert component_registered("fooponent")
    assert viscosity_available("fooponent")
    assert thermal_conductivity_available("fooponent")
    assert surface_tension_available("fooponent")
    assert eos_reference("fooponent") == "cite eos"
    assert viscosity_reference("fooponent") == "cite viscosity"
    assert thermal_conductivity_reference("fooponent") == "cite thermal conductivity"
    assert surface_tension_reference("fooponent") == "cite surface tension"
    remove_component("fooponent")
    assert not component_registered("fooponent")


@pytest.mark.unit
def test_register_and_get_list_refs():
    register_helmholtz_component(
        "fooponent",
        viscosity=True,
        thermal_conductivity=True,
        surface_tension=True,
        eos_ref=["cite", "eos"],
        viscosity_ref=["cite", "viscosity"],
        thermal_conductivity_ref=["cite", "thermal conductivity"],
        surface_tension_ref=["cite", "surface tension"],
    )
    assert component_registered("fooponent")
    assert viscosity_available("fooponent")
    assert thermal_conductivity_available("fooponent")
    assert surface_tension_available("fooponent")
    assert eos_reference("fooponent") == "cite\neos"
    assert viscosity_reference("fooponent") == "cite\nviscosity"
    assert thermal_conductivity_reference("fooponent") == "cite\nthermal conductivity"
    assert surface_tension_reference("fooponent") == "cite\nsurface tension"
    remove_component("fooponent")
    assert not component_registered("fooponent")


@pytest.mark.unit
def test_not_registered_get():
    assert not component_registered("fooponent")
    assert not viscosity_available("fooponent")
    assert not thermal_conductivity_available("fooponent")
    assert not surface_tension_available("fooponent")
    assert eos_reference("fooponent") is None
    assert viscosity_reference("fooponent") is None
    assert thermal_conductivity_reference("fooponent") is None
    assert surface_tension_reference("fooponent") is None



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Method to set constant pure component properties:

"""
# TODO: Missing doc strings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.environ import log, Var

from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
# Heat capacities, enthalpies and entropies
class Constant(object):

    # Ideal liquid properties methods
    class cp_mol_liq_comp(object):
        @staticmethod
        def build_parameters(cobj):
            units = cobj.parent_block().get_metadata().derived_units
            cobj.cp_mol_liq_comp_coeff = Var(
                doc="Parameter for liquid phase molar heat capacity",
                units=units.HEAT_CAPACITY_MOLE,
            )
            set_param_from_config(cobj, param="cp_mol_liq_comp_coeff")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific heat capacity
            return cobj.cp_mol_liq_comp_coeff

    class enth_mol_liq_comp(object):
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_liq_comp_coeff"):
                Constant.cp_mol_liq_comp.build_parameters(cobj)

            if cobj.parent_block().config.include_enthalpy_of_formation:
                units = cobj.parent_block().get_metadata().derived_units

                cobj.enth_mol_form_liq_comp_ref = Var(
                    doc="Liquid phase molar heat of formation @ Tref",
                    units=units.ENERGY_MOLE,
                )
                set_param_from_config(cobj, param="enth_mol_form_liq_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific enthalpy
            units = b.params.get_metadata().derived_units
            Tr = b.params.temperature_ref

            h_form = (
                cobj.enth_mol_form_liq_comp_ref
                if b.params.config.include_enthalpy_of_formation
                else 0 * units.ENERGY_MOLE
            )

            return cobj.cp_mol_liq_comp_coeff * (T - Tr) + h_form

    class entr_mol_liq_comp(object):
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_liq_comp_coeff"):
                Constant.cp_mol_liq_comp.build_parameters(cobj)

            units = cobj.parent_block().get_metadata().derived_units

            cobj.entr_mol_form_liq_comp_ref = Var(
                doc="Liquid phase molar entropy of formation @ Tref",
                units=units.ENTROPY_MOLE,
            )
            set_param_from_config(cobj, param="entr_mol_form_liq_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific entropy
            Tr = b.params.temperature_ref

            return (
                cobj.cp_mol_liq_comp_coeff * log(T / Tr)
                + cobj.entr_mol_form_liq_comp_ref
            )

    class dens_mol_liq_comp(object):
        @staticmethod
        def build_parameters(cobj):
            units = cobj.parent_block().get_metadata().derived_units
            cobj.dens_mol_liq_comp_coeff = Var(
                doc="Parameter for liquid phase molar density",
                units=units.DENSITY_MOLE,
            )
            set_param_from_config(cobj, param="dens_mol_liq_comp_coeff")

        @staticmethod
        def return_expression(b, cobj, T):
            # Molar density
            return cobj.dens_mol_liq_comp_coeff

    # Ideal gas properties methods
    class cp_mol_ig_comp(object):
        @staticmethod
        def build_parameters(cobj):
            units = cobj.parent_block().get_metadata().derived_units
            cobj.cp_mol_ig_comp_coeff = Var(
                doc="Parameter for ideal gas molar heat capacity",
                units=units.HEAT_CAPACITY_MOLE,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific heat capacity
            return cobj.cp_mol_ig_comp_coeff

    class enth_mol_ig_comp(object):
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff"):
                Constant.cp_mol_ig_comp.build_parameters(cobj)

            if cobj.parent_block().config.include_enthalpy_of_formation:
                units = cobj.parent_block().get_metadata().derived_units

                cobj.enth_mol_form_ig_comp_ref = Var(
                    doc="Ideal gas molar heat of formation @ Tref",
                    units=units.ENERGY_MOLE,
                )
                set_param_from_config(cobj, param="enth_mol_form_ig_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific enthalpy
            units = b.params.get_metadata().derived_units
            Tr = b.params.temperature_ref

            h_form = (
                cobj.enth_mol_form_ig_comp_ref
                if b.params.config.include_enthalpy_of_formation
                else 0 * units.ENERGY_MOLE
            )

            return cobj.cp_mol_ig_comp_coeff * (T - Tr) + h_form

    class entr_mol_ig_comp(object):
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff"):
                Constant.cp_mol_ig_comp.build_parameters(cobj)

            units = cobj.parent_block().get_metadata().derived_units

            cobj.entr_mol_form_ig_comp_ref = Var(
                doc="Ideal gas molar entropy of formation @ Tref",
                units=units.ENTROPY_MOLE,
            )
            set_param_from_config(cobj, param="entr_mol_form_ig_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific entropy
            Tr = b.params.temperature_ref

            return (
                cobj.cp_mol_ig_comp_coeff * log(T / Tr) + cobj.entr_mol_form_ig_comp_ref
            )

    # Ideal solid properties methods
    class cp_mol_sol_comp(object):
        @staticmethod
        def build_parameters(cobj):
            units = cobj.parent_block().get_metadata().derived_units
            cobj.cp_mol_sol_comp_coeff = Var(
                doc="Parameter for solid phase molar heat capacity",
                units=units.HEAT_CAPACITY_MOLE,
            )
            set_param_from_config(cobj, param="cp_mol_sol_comp_coeff")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific heat capacity
            cp = cobj.cp_mol_sol_comp_coeff
            return cp

    class enth_mol_sol_comp(object):
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_sol_comp_coeff"):
                Constant.cp_mol_sol_comp.build_parameters(cobj)

            if cobj.parent_block().config.include_enthalpy_of_formation:
                units = cobj.parent_block().get_metadata().derived_units

                cobj.enth_mol_form_sol_comp_ref = Var(
                    doc="Solid phase molar heat of formation @ Tref",
                    units=units.ENERGY_MOLE,
                )
                set_param_from_config(cobj, param="enth_mol_form_sol_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific enthalpy
            units = b.params.get_metadata().derived_units
            Tr = b.params.temperature_ref

            h_form = (
                cobj.enth_mol_form_sol_comp_ref
                if b.params.config.include_enthalpy_of_formation
                else 0 * units.ENERGY_MOLE
            )

            return cobj.cp_mol_sol_comp_coeff * (T - Tr) + h_form

    class entr_mol_sol_comp(object):
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_sol_comp_coeff"):
                Constant.cp_mol_sol_comp.build_parameters(cobj)

            units = cobj.parent_block().get_metadata().derived_units

            cobj.entr_mol_form_sol_comp_ref = Var(
                doc="Solid phase molar entropy of formation @ Tref",
                units=units.ENTROPY_MOLE,
            )
            set_param_from_config(cobj, param="entr_mol_form_sol_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific entropy
            Tr = b.params.temperature_ref

            return (
                cobj.cp_mol_sol_comp_coeff * log(T / Tr)
                + cobj.entr_mol_form_sol_comp_ref
            )

    class dens_mol_sol_comp(object):
        @staticmethod
        def build_parameters(cobj):
            units = cobj.parent_block().get_metadata().derived_units
            cobj.dens_mol_sol_comp_coeff = Var(
                doc="Parameter for solid phase molar density",
                units=units.DENSITY_MOLE,
            )
            set_param_from_config(cobj, param="dens_mol_sol_comp_coeff")

        @staticmethod
        def return_expression(b, cobj, T):
            # Molar density
            return cobj.dens_mol_sol_comp_coeff

    class visc_d_phase_comp(object):
        @staticmethod
        def build_parameters(cobj, p):
            units = cobj.parent_block().get_metadata().derived_units
            # Calling this a "coefficient" doesn't make much sense, but want to be consistent with other methods
            cobj.add_component(
                f"visc_d_{p}_comp_coeff",
                Var(
                    doc=f"Parameter for {p} phase dynamic viscosity",
                    units=units["dynamic_viscosity"],
                ),
            )
            set_param_from_config(cobj, param=f"visc_d_{p}_comp_coeff")

        @staticmethod
        def return_expression(b, cobj, p, T):
            return getattr(cobj, f"visc_d_{p}_comp_coeff")

    class therm_cond_phase_comp(object):
        @staticmethod
        def build_parameters(cobj, p):
            units = cobj.parent_block().get_metadata().derived_units
            cobj.add_component(
                f"therm_cond_{p}_comp_coeff",
                Var(
                    doc=f"Parameter for {p} phase thermal conductivity",
                    units=units["thermal_conductivity"],
                ),
            )
            set_param_from_config(cobj, param=f"therm_cond_{p}_comp_coeff")

        @staticmethod
        def return_expression(b, cobj, p, T):
            return getattr(cobj, f"therm_cond_{p}_comp_coeff")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for MEA solvent column model
Author: Anuja Deshpande, Andrew Lee
"""
# Import Python libraries
import copy
import pytest

# Import Pyomo libraries
from pyomo.environ import (
    ConcreteModel,
    value,
    assert_optimal_termination,
    units as pyunits,
)
import pyomo.common.unittest as unittest

# Import IDAES Libraries
import idaes
from idaes.core import FlowsheetBlock
from idaes.models_extra.column_models.MEAsolvent_column import MEAColumn
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models_extra.column_models.properties.MEA_vapor import (
    flue_gas as vaporconfig,
)
from idaes.models_extra.column_models.properties.MEA_solvent import (
    configuration as liquidconfig,
)

from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver
from idaes.core.util.performance import PerformanceBaseClass

import idaes.logger as idaeslog

solver = get_solver()


# -----------------------------------------------------------------------------
def build_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # Set up molar flow bounds
    vaporconfig_copy = copy.deepcopy(vaporconfig)
    liquidconfig_copy = copy.deepcopy(liquidconfig)
    vaporconfig_copy["state_bounds"]["flow_mol"] = (
        0,
        1,
        1e6,
        pyunits.mol / pyunits.s,
    )
    liquidconfig_copy["state_bounds"]["flow_mol"] = (
        0,
        1,
        1e6,
        pyunits.mol / pyunits.s,
    )

    # Set up property package
    m.fs.vapor_properties = GenericParameterBlock(**vaporconfig_copy)
    m.fs.liquid_properties = GenericParameterBlock(**liquidconfig_copy)

    # Number of finite elements and finite element list in the spatial domain
    x_nfe = 40

    # Create an instance of the column in the flowsheet
    m.fs.unit = MEAColumn(
        finite_elements=x_nfe,
        vapor_phase={"property_package": m.fs.vapor_properties},
        liquid_phase={"property_package": m.fs.liquid_properties},
    )

    # Fix column design variables
    # Absorber diameter
    m.fs.unit.diameter_column.fix(18)

    # Absorber length
    m.fs.unit.length_column.fix(21.6)  # meter

    # Fix operating conditions

    # Flue gas inlet
    m.fs.unit.vapor_inlet.flow_mol.fix(19000)  # mol/sec
    m.fs.unit.vapor_inlet.temperature.fix(313.15)  # K
    m.fs.unit.vapor_inlet.pressure.fix(100000)  # Pa
    m.fs.unit.vapor_inlet.mole_frac_comp[0, "CO2"].fix(0.041)
    m.fs.unit.vapor_inlet.mole_frac_comp[0, "H2O"].fix(0.074)
    m.fs.unit.vapor_inlet.mole_frac_comp[0, "N2"].fix(0.763)
    m.fs.unit.vapor_inlet.mole_frac_comp[0, "O2"].fix(0.122)

    # Lean solvent inlet
    m.fs.unit.liquid_inlet.flow_mol.fix(24000)  # mol/sec
    m.fs.unit.liquid_inlet.temperature.fix(313.15)  # K
    m.fs.unit.liquid_inlet.pressure.fix(100000)  # Pa
    m.fs.unit.liquid_inlet.mole_frac_comp[0, "CO2"].fix(0.022)
    m.fs.unit.liquid_inlet.mole_frac_comp[0, "H2O"].fix(0.868)
    m.fs.unit.liquid_inlet.mole_frac_comp[0, "MEA"].fix(0.11)

    return m


@pytest.mark.performance
class Test_MEAColumn_Performance(PerformanceBaseClass, unittest.TestCase):
    def build_model(self):
        return build_model()

    def initialize_model(self, model):
        model.fs.unit.initialize()

    def solve_model(self, model):
        with idaes.temporary_config_ctx():
            # Get default solver for testing
            solver.options["bound_push"] = 1e-10

            res = solver.solve(model)

            assert_optimal_termination(res)


class TestAbsorber:
    @pytest.fixture(scope="class")
    def model(self):
        return build_model()

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.component
    @pytest.mark.xfail  # TODO: Remove once model is fixed
    def test_initialize(self, model):
        initialization_tester(model, outlvl=idaeslog.DEBUG)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    @pytest.mark.xfail  # TODO: Remove once model is fixed
    def test_solve(self, model):
        with idaes.temporary_config_ctx():
            # Get default solver for testing
            solver.options["bound_push"] = 1e-10

            res = solver.solve(model)

        # Solver status/condition
        assert_optimal_termination(res)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    @pytest.mark.xfail  # TODO: Remove once model is fixed
    def test_solution(self, model):
        assert pytest.approx(19436.448618, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.flow_mol[0]
        )
        assert pytest.approx(0.00127033, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.1336025, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.74586671, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "N2"]
        )
        assert pytest.approx(0.11926047, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.mole_frac_comp[0, "O2"]
        )
        assert pytest.approx(100000.0, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.pressure[0]
        )
        assert pytest.approx(327.03911, rel=1e-5) == value(
            model.fs.unit.vapor_outlet.temperature[0]
        )

        assert pytest.approx(23563.551381, rel=1e-5) == value(
            model.fs.unit.liquid_outlet.flow_mol[0]
        )
        assert pytest.approx(0.05441918, rel=1e-5) == value(
            model.fs.unit.liquid_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.83354337, rel=1e-5) == value(
            model.fs.unit.liquid_outlet.mole_frac_comp[0, "H2O"]
        )
        assert pytest.approx(0.11203744, rel=1e-5) == value(
            model.fs.unit.liquid_outlet.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(100000.0, rel=1e-5) == value(
            model.fs.unit.liquid_outlet.pressure[0]
        )
        assert pytest.approx(315.27424489, rel=1e-5) == value(
            model.fs.unit.liquid_outlet.temperature[0]
        )

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        vap_in = model.fs.unit.vapor_phase.properties[0, 0]
        vap_out = model.fs.unit.vapor_phase.properties[0, 1]
        liq_in = model.fs.unit.liquid_phase.properties[0, 1]
        liq_out = model.fs.unit.liquid_phase.properties[0, 0]

        # Material conservation
        for j in ["CO2", "H2O"]:
            assert 1e-6 >= abs(
                value(
                    vap_in.get_material_flow_terms("Vap", j)
                    + liq_in.get_material_flow_terms("Liq", j)
                    - vap_out.get_material_flow_terms("Vap", j)
                    - liq_out.get_material_flow_terms("Liq", j)
                )
            )
        for j in ["N2", "O2"]:
            assert 1e-6 >= abs(
                value(
                    vap_in.get_material_flow_terms("Vap", j)
                    - vap_out.get_material_flow_terms("Vap", j)
                )
            )
        for j in ["MEA"]:
            assert 1e-6 >= abs(
                value(
                    liq_in.get_material_flow_terms("Liq", j)
                    - liq_out.get_material_flow_terms("Liq", j)
                )
            )

        # Energy conservation
        assert 1e-6 >= abs(
            value(
                vap_in.get_enthalpy_flow_terms("Vap")
                + liq_in.get_enthalpy_flow_terms("Liq")
                - vap_out.get_enthalpy_flow_terms("Vap")
                - liq_out.get_enthalpy_flow_terms("Liq")
            )
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pyomo.environ as pyo
import os
from idaes.apps.grid_integration import Tracker
from idaes.apps.grid_integration import Bidder
from idaes.apps.grid_integration import PlaceHolderForecaster
from idaes.apps.grid_integration import DoubleLoopCoordinator
from idaes.apps.grid_integration.examples.thermal_generator import ThermalGenerator
from idaes.apps.grid_integration.examples.utils import (
    rts_gmlc_generator_dataframe,
    rts_gmlc_bus_dataframe,
    daily_da_price_means,
    daily_rt_price_means,
    daily_da_price_stds,
    daily_rt_price_stds,
)

this_module_dir = os.path.dirname(__file__)

generator = "10_STEAM"
tracking_horizon = 4
day_ahead_bidding_horizon = 48
real_time_bidding_horizon = tracking_horizon
n_scenario = 10
n_tracking_hour = 1

# create forecaster
forecaster = PlaceHolderForecaster(
    daily_da_price_means=daily_da_price_means,
    daily_rt_price_means=daily_rt_price_means,
    daily_da_price_stds=daily_da_price_stds,
    daily_rt_price_stds=daily_rt_price_stds,
)

# create solver
solver = pyo.SolverFactory("cbc")

## create trackers

# make a tracker
tracking_model_object = ThermalGenerator(
    rts_gmlc_generator_dataframe=rts_gmlc_generator_dataframe,
    rts_gmlc_bus_dataframe=rts_gmlc_bus_dataframe,
    generator=generator,
)
thermal_tracker = Tracker(
    tracking_model_object=tracking_model_object,
    tracking_horizon=tracking_horizon,
    n_tracking_hour=n_tracking_hour,
    solver=solver,
)

# make a projection tracker
projection_tracking_model_object = ThermalGenerator(
    rts_gmlc_generator_dataframe=rts_gmlc_generator_dataframe,
    rts_gmlc_bus_dataframe=rts_gmlc_bus_dataframe,
    generator=generator,
)
thermal_projection_tracker = Tracker(
    tracking_model_object=projection_tracking_model_object,
    tracking_horizon=tracking_horizon,
    n_tracking_hour=n_tracking_hour,
    solver=solver,
)

# create a bidder
bidding_model_object = ThermalGenerator(
    rts_gmlc_generator_dataframe=rts_gmlc_generator_dataframe,
    rts_gmlc_bus_dataframe=rts_gmlc_bus_dataframe,
    generator=generator,
)
thermal_bidder = Bidder(
    bidding_model_object=bidding_model_object,
    day_ahead_horizon=day_ahead_bidding_horizon,
    real_time_horizon=real_time_bidding_horizon,
    n_scenario=n_scenario,
    solver=solver,
    forecaster=forecaster,
)

# create coordinator
coordinator = DoubleLoopCoordinator(
    bidder=thermal_bidder,
    tracker=thermal_tracker,
    projection_tracker=thermal_projection_tracker,
)

## Prescient requires the following functions in this module
get_configuration = coordinator.get_configuration
register_plugins = coordinator.register_plugins



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Test the Helmholtz EoS parameter writing utility.  Property calculations 
tested here do not use IDAES properties or the external functions. This directly
tests the expressions created by the WriteParameters class that will be written 
to parameter and expression files to verify they are generated correctly.

The tests here use three points from the saturation curve near critical, near triple,
and in between.  This data comes from the original equation of state papers.  Comments
include citations.

On some test points, I had to infer extra significant figures for density.  For example,
this density of a liquid near the triple point is insensitive to pressure, so to calculate
the pressure from density of a liquid at the triple point, a lot of significant figures 
are needed. As long a I could add significant figures to the density and not change the 
reported value when rounded, I assumed it was okay. 
"""

import pytest
from idaes.models.properties.general_helmholtz.components.parameters.h2o import (
    main as h2o_main,
)
from idaes.models.properties.general_helmholtz.components.parameters.co2 import (
    main as co2_main,
)
from idaes.models.properties.general_helmholtz.components.parameters.propane import (
    main as propane_main,
)
from idaes.models.properties.general_helmholtz.components.parameters.r1234ze import (
    main as r1234ze_main,
)
from idaes.models.properties.general_helmholtz.components.parameters.r125 import (
    main as r125_main,
)
from idaes.models.properties.general_helmholtz.components.parameters.r134a import (
    main as r134a_main,
)
from idaes.models.properties.general_helmholtz.components.parameters.r227ea import (
    main as r227ea_main,
)
from idaes.models.properties.general_helmholtz.components.parameters.r32 import (
    main as r32_main,
)


def _common_sat(sat_thermo_data, we):
    # Check the EoS expressions, the tolerance is a little loose
    # due to lack of sig. figs. in reported data
    for pnt in sat_thermo_data.values():
        assert we.calculate_pressure(rho=pnt["rhov"], T=pnt["T"]) == pytest.approx(
            pnt["p"], rel=1e-2, abs=1e-3
        )
        assert we.calculate_pressure(rho=pnt["rhol"], T=pnt["T"]) == pytest.approx(
            pnt["p"], rel=1e-2, abs=1e-3
        )
        assert we.calculate_enthalpy(rho=pnt["rhov"], T=pnt["T"]) == pytest.approx(
            pnt["hv"], rel=1e-2, abs=1e-3
        )
        assert we.calculate_enthalpy(rho=pnt["rhol"], T=pnt["T"]) == pytest.approx(
            pnt["hl"], rel=1e-2, abs=1e-3
        )
        assert we.calculate_entropy(rho=pnt["rhov"], T=pnt["T"]) == pytest.approx(
            pnt["sv"], rel=1e-2, abs=1e-3
        )
        assert we.calculate_entropy(rho=pnt["rhol"], T=pnt["T"]) == pytest.approx(
            pnt["sl"], rel=1e-2, abs=1e-3
        )

    # Check the approximate sat density curves
    rhol, rhov = we.approx_sat_curves([sat_thermo_data[2]["T"]])
    assert rhol[0] == pytest.approx(sat_thermo_data[2]["rhol"], rel=1e-1)
    assert rhov[0] == pytest.approx(sat_thermo_data[2]["rhov"], rel=1e-1)


def _common_pressure(thermo_data, we):
    # Check the EoS expressions, the tolerance is a little loose
    # due to lack of sig. figs. in reported data
    for pnt in thermo_data.values():
        assert we.calculate_pressure(rho=pnt["rho"], T=pnt["T"]) == pytest.approx(
            pnt["p"], rel=1e-2, abs=1e-3
        )


@pytest.mark.unit
def test_h2o():
    # Some test data from:
    #
    # Wagner, W.,  A. Pruss (2002). The IAPWS Formulation 1995 for the
    #   Thermodynamic Properties of Ordinary Water Substance for General and
    #   Scientific Use. J. Phys. Chem. Ref. Data, 31, 387-535.
    sat_thermo_data = {
        1: {  # near critical
            "T": 273.16,
            "p": 0.612,
            "rhol": 999.79252,
            "hl": 0.001,
            "sl": 0.0000,
            "rhov": 0.00485,
            "hv": 2500.92,
            "sv": 9.1555,
        },
        2: {  # between critical and triple point
            "T": 400,
            "p": 245.77,
            "rhol": 937.486,
            "hl": 532.953,
            "sl": 1.6013,
            "rhov": 1.3694,
            "hv": 2715.70,
            "sv": 7.0581,
        },
        3: {  # near triple point
            "T": 646,
            "p": 21775,
            "rhol": 402.96,
            "hl": 1963.49,
            "sl": 4.2214,
            "rhov": 243.46,
            "hv": 2238.06,
            "sv": 4.6465,
        },
    }

    we = h2o_main(dry_run=True)
    _common_sat(sat_thermo_data, we)


@pytest.mark.unit
def test_co2():
    # Some test data from:
    #
    # Span, R., and W. Wagner (1996). A New Equation of State for Carbon Dioxide
    #    Covering the Fluid Region from the Triple-Point Temperature to 1100 K as
    #    Pressures up to 800 MPa. Journal of Physical and Chemical Reference Data,
    #    25, 1509.
    sat_thermo_data = {
        1: {  # near critical
            "T": 216.592,
            "p": 517.96,
            "rhol": 1178.46,
            "hl": -426.74,
            "sl": -2.2177,
            "rhov": 13.761,
            "hv": -76.364,
            "sv": -0.59999,
        },
        2: {  # between critical and triple point
            "T": 250,
            "p": 1785.0,
            "rhol": 1045.97,
            "hl": -359.07,
            "sl": -1.9323,
            "rhov": 46.644,
            "hv": -69.376,
            "sv": -0.77492,
        },
        3: {  # near triple point
            "T": 303,
            "p": 7189.0,
            "rhol": 599.86,
            "hl": -203.73,
            "sl": -1.4004,
            "rhov": 339.00,
            "hv": -139.91,
            "sv": -1.1897,
        },
    }

    we = co2_main(dry_run=True)
    _common_sat(sat_thermo_data, we)


@pytest.mark.unit
def test_propane():
    # Some test data from:
    #
    # Lemmon, E. , McLinden, M. and Wagner, W. (2009), Thermodynamic Properties of
    #   Propane. III. A Reference Equation of State for Temperatures from the Melting
    #   Line to 650 K and Pressures up to 1000 MPa, Journal of Chemical and Engineering
    #   Data, [online], https://tsapps.nist.gov/publication/get_pdf.cfm?pub_id=832438
    #   (Accessed April 14, 2023)
    sat_thermo_data = {
        1: {  # near triple
            "T": -187.625 + 273.15,
            "p": 0.17203e-6,
            "rhol": 733.125204,
            "hl": -196.64,
            "sl": -1.396,
            "rhov": 0.1071e-7,
            "hv": 366.26,
            "sv": 5.186,
        },
        2: {  # between critical and triple point
            "T": 273.15,
            "p": 474.46,
            "rhol": 528.59,
            "hl": 200.00,
            "sl": 1.000,
            "rhov": 10.351,
            "hv": 574.87,
            "sv": 2.372,
        },
        3: {  # near critical point
            "T": 95 + 273.15,
            "p": 4119.5,
            "rhol": 286.51,
            "hl": 516.33,
            "sl": 1.948,
            "rhov": 156.31,
            "hv": 595.81,
            "sv": 2.164,
        },
    }

    we = propane_main(dry_run=True)
    _common_sat(sat_thermo_data, we)


@pytest.mark.unit
def test_r1234ze():
    # Some test data from:
    #
    # Monika Thol and Eric W. Lemmon. Equation of State for the Thermodynamic
    #    Properties of trans-1,3,3,3-Tetrafluoropropene [R-1234ze(E)]. Int. J.
    #    Thermophys, 37(3):1–16, 2016. doi:10.1007/s10765-016-2040-6.
    thermo_data = {
        1: {
            "T": 200,
            "p": 2161.4,
            "rho": 12.6 * 114.0416,
        },
        2: {
            "T": 200,
            "p": 0,
            "rho": 1e-12,  # Need a small number here to avoid 0/0
        },
        3: {
            "T": 350,
            "p": 95041.56,
            "rho": 11.4 * 114.0416,
        },
        4: {
            "T": 360,
            "p": 2039.103,
            "rho": 1.0 * 114.0416,
        },
        4: {
            "T": 383,
            "p": 3670.105,
            "rho": 4.29 * 114.0416,
        },
        5: {
            "T": 420,
            "p": 19954.47,
            "rho": 8.0 * 114.0416,
        },
    }

    we = r1234ze_main(dry_run=True)
    _common_pressure(thermo_data, we)


@pytest.mark.unit
def test_r125():
    # Some test data from:
    #
    # Lemmon, E.W., R.T. Jacobsen. A New Functional Form and New Fitting Techniques
    #    for Equations of State with Application to Pentafluoroethane (HFC-125),
    #    J. Phys. Chem. Ref. Data, Vol. 34, No. 1, 2005
    sat_thermo_data = {
        1: {  # near triple point
            "T": -100.63 + 273.15,
            "p": 2.91,
            "rhol": 1690.6804,
            "hl": 87.130,
            "sl": 0.49022,
            "rhov": 0.24462,
            "hv": 277.39,
            "sv": 1.5931,
        },
        2: {  # between critical and triple point
            "T": 273.15,
            "p": 670.52,
            "rhol": 1319.8,
            "hl": 200.0,
            "sl": 1.000,
            "rhov": 42.070,
            "hv": 333.16,
            "sv": 1.4875,
        },
        3: {  # near critical point
            "T": 65 + 273.15,
            "p": 3536.97,
            "rhol": 735.11,
            "hl": 304.88,
            "sl": 1.3311,
            "rhov": 416.57,
            "hv": 332.24,
            "sv": 1.412,
        },
    }

    we = r125_main(dry_run=True)
    _common_sat(sat_thermo_data, we)


@pytest.mark.unit
def test_r134a():
    # Some test data from:
    #
    # Tillner-Roth, R.; Baehr, H.D., An International Standard Formulation for the",
    #    Thermodynamic Properties of 1,1,1,2-Tetrafluoroethane (HFC-134a) for",
    #    Temperatures from 170 K to 455 K and Pressures up to 70 MPa, J. Phys. Chem.",
    #    Ref. Data, 1994, 23, 5, 657-729, https://doi.org/10.1063/1.555958"
    sat_thermo_data = {
        1: {  # near triple point
            "T": 169.85,
            "p": 0.39,
            "rhol": 1591.10745,
            "hl": 71.454,
            "sl": 0.4126,
            "rhov": 0.02817,
            "hv": 334.94,
            "sv": 1.9639,
        },
        2: {  # between critical and triple point
            "T": 310.0,
            "p": 933.40,
            "rhol": 1159.9,
            "hl": 251.73,
            "sl": 1.1756,
            "rhov": 45.785,
            "hv": 418.03,
            "sv": 1.7121,
        },
        3: {  # near critical point
            "T": 372,
            "p": 3881.1,
            "rhol": 693.10,
            "hl": 367.71,
            "sl": 1.5041,
            "rhov": 334.83,
            "hv": 412.67,
            "sv": 1.6250,
        },
    }

    we = r134a_main(dry_run=True)
    _common_sat(sat_thermo_data, we)


@pytest.mark.unit
def test_r32():
    # Some test data from:
    #
    # Tillner-Roth, R., A. Yokozeki. An International Standard Equation of State for",
    #    Difluoromethane (R-32) for Temperatures from the Triple Point at 136.34 K to",
    #    435 K and Pressures up to 70 MPa, Journal of Physical and Chemical Reference",
    #    Data 26, 1273 (1997); https://doi.org/10.1063/1.556002"
    sat_thermo_data = {
        1: {  # near triple point
            "T": -136.81 + 273.15,
            "p": 0.05446079,
            "rhol": 1429.2733,
            "hl": -19.07,
            "sl": -0.104,
            "rhov": 0.0025,
            "hv": 444.31,
            "sv": 3.2937,
        },
        2: {  # between critical and triple point
            "T": 273.15,
            "p": 813.10,
            "rhol": 1055.25,
            "hl": 200.00,
            "sl": 1.0,
            "rhov": 22.091,
            "hv": 515.30,
            "sv": 2.1543,
        },
        3: {  # near critical point
            "T": 76 + 273.15,
            "p": 5531.5,
            "rhol": 583.33,
            "hl": 378.03,
            "sl": 1.5470,
            "rhov": 275.00,
            "hv": 455.86,
            "sv": 1.7699,
        },
    }

    we = r32_main(dry_run=True)
    _common_sat(sat_thermo_data, we)
